#!/usr/bin/env python

import pygame
import os
import random
import pdb
from pygame.locals import *

screen_size = (900, 300)
block_pixels = (30, 30)
block_dimensions = tuple([i / j for i, j in zip(screen_size, block_pixels)])
DISPLAY_REFRESH = USEREVENT
FPS = 40

def post_quit_event():
    """Posts a quit event to the python event queue"""
    pygame.event.post(pygame.event.Event(pygame.QUIT))


def post_refresh_event():
    """Posts a refresh event to the python event queue"""
    pygame.event.post(pygame.event.Event(DISPLAY_REFRESH))


def main():
    os.environ['SDL_VIDEO_WINDOW_POS'] = 'center'
    pygame.init()
    screen = pygame.display.set_mode(screen_size)
    game(screen)


def game(screen):
    Level(400, 15).play(screen)
    
    
class Block(object):
    def __init__(self, color):
        self.color = color

    def draw(self, screen, position):
        if self.color:
            screen.fill(self.color, pygame.Rect(position, block_pixels))

    def isEmpty(self):
        return self.color == None
        
    def __eq__(self, obj):
        if type(self) != type(obj):
            return False
        else:
            return self.color == obj.color

    def __ne__(self, obj):
        return not self == obj

    def __str__(self):
        return str(self.color)
        
    def __repr__(self):
        return repr(self.color)

SENTINEL = Block(None)    


class Wall(object):
    def __init__(self, width):
        self.width = width
        self.array = self._generate_wall(width)
        self.highlight = None
        
    def set_highlight(self, x, y):
        self.highlight = (x, y, (0, 0, 0))

    def set_selected(self, x, y):
        self.highlight = (x, y, (255, 0, 0))
        
    def clear_highlight(self):
        self.highlight = None

    def isEmpty(self, x, y):
        return self.array[x][y].isEmpty()

    def swap(self, (ax, ay), (bx, by)):
        temp = self.array[bx][by]
        self.array[bx][by] = self.array[ax][ay]
        self.array[ax][ay] = temp
        self.remove_pieces(ax, ay)
        self.remove_pieces(bx, by)
        
    def remove_pieces(self, x, y):
        def child_list(pos, previous):
            parent = self.array[pos[0]][pos[1]]
            def isChild(pos):
                x, y = pos
                if (x < 0 or x >= self.width or
                    y < 0 or y >= block_dimensions[1] or
                    (x, y) in previous or parent != self.array[x][y]):
                    return False
                else:
                    return True
            x, y = pos
            a = (x + 1, y)
            b = (x - 1, y)
            c = (x, y + 1)
            d = (x, y - 1)
            return [i for i in (a, b, c, d) if isChild(i)]

        def count(pos, previous):
            previous.update([pos])
            return 1 + sum([count(i, previous) for i in child_list(pos, previous)])
            
        def remove(pos, previous):
            x, y = pos
            previous.update([pos])
            for child in child_list(pos, previous):
                remove(child, previous)
            self.array[x][y] = SENTINEL

        if count((x, y), set()) > 2:
            remove((x, y), set())

    def draw(self, screen, xpos):
        if xpos >= 0:
            xblock = xpos / block_dimensions[0]
            xoff = xpos % block_dimensions[0]
        else:
            xblock = -(abs(xpos) / block_dimensions[0])
            xoff = -(abs(xpos) % block_dimensions[0])
        for i in range(self.width):
            for j in range(block_dimensions[1]):
                x = (i + xblock) * block_pixels[0] +  xoff
                y = j * block_pixels[1]
                self.array[i][j].draw(screen, (x, y))
        if self.highlight:
            x, y, color = self.highlight
            if not self.array[x][y].isEmpty():
                x = (x + xblock) * block_pixels[0] +  xoff
                y = y * block_pixels[1]
                pygame.draw.rect(screen, color, ((x, y), block_pixels), 4)                

    @staticmethod
    def _generate_wall(width):
        pieces = (Block(pygame.Color('green')),
                  Block(pygame.Color('pink')),
                  Block(pygame.Color('orange')),
                  Block(pygame.Color('violet')))
        wall = list()
        for i in range(width):
            column = list()
            for j in range(block_dimensions[1]):
                piece = random.choice(pieces)
                if i > 1:
                    while piece == wall[i - 1][j] and piece == wall[i - 2][j]:
                        piece = random.choice(pieces)
                column.append(piece)
            wall.append(column)
        return wall
        

class Level(object):
    def __init__(self, width, speed):
        self.width = width
        self.wall = Wall(self.width)
        self.speed = speed
        self.start = None
        self.offset = screen_size[0]
        self.mpos = None
        self.water = [[False for i in range(block_dimensions[1])] for j in range(2 * block_dimensions[0] + width)]
        for i in range(block_dimensions[0] + 1):
            for j in range(block_dimensions[1]):
                self.water[i][j] = True
        
    def play(self, screen):
        self.start = pygame.time.get_ticks()
        post_refresh_event()
        pygame.time.set_timer(DISPLAY_REFRESH, int(1000.0 / FPS))
        running = True
        while running:
            for evt in [pygame.event.wait()] + pygame.event.get():
                if evt.type == pygame.QUIT:
                    running = False
                elif evt.type == DISPLAY_REFRESH:
                    self.update_mouse_pos(pygame.mouse.get_pos())
                    self.draw_background(screen)
                    self.draw_stage(screen)
                    self.update_mouse_pos(pygame.mouse.get_pos())
                    #self.draw_water(screen)
                    pygame.display.flip()
                elif evt.type == pygame.MOUSEMOTION:
                    if not (self.mpos or self.isInvalidPosition(evt.pos)):
                        self.update_mouse_pos(evt.pos)
                elif evt.type == pygame.MOUSEBUTTONDOWN:
                    if evt.button == 1:
                        self.press_mouse(evt.pos)
                elif evt.type == pygame.MOUSEBUTTONUP:
                    if evt.button == 1:
                        self.release_mouse(evt.pos)

    def release_mouse(self, pos, clear=True):
        mx, my = self.relativepos(pos)
        if not self.mpos or self.isInvalidPosition(pos) or (mx, my) == self.mpos:
            self.clearMouse()
            return
        if mx == self.mpos[0]:
            start = self.mpos[1]
            end = my
            step = 1 if end > start else -1
            while start != end:
                if self.wall.isEmpty(mx, start + step):
                    break
                self.wall.swap((mx, start), (mx, start + step))
                start += step
        if my == self.mpos[1]:
            start = self.mpos[0]
            end = mx
            step = 1 if end > start else -1
            while start != end:
                if self.wall.isEmpty(start + step, my):
                    break
                self.wall.swap((start, my), (start + step, my))
                start += step
        if clear:
            self.clearMouse()
                    
    def press_mouse(self, pos):
        if self.isInvalidPosition(pos):
            self.clearMouse()
            return
        else:
            nx, ny = self.relativepos(pos)
            self.mpos = nx, ny
            self.wall.set_selected(nx, ny)
           
    def isInvalidPosition(self, pos):
        mx, my = pos
        if (mx < self.offset or mx >= self.offset +
                self.width * block_pixels[0] or
                my < 0 or my >= screen_size[1]):
            return False
        nx, ny = self.relativepos(pos)
        return self.wall.isEmpty(nx, ny)
        
    def clearMouse(self):
        self.mpos = None
        self.wall.clear_highlight()
        
    def relativepos(self, pos):
        mx, my = pos
        nx = (mx - self.offset) / block_pixels[0]
        ny = my / block_pixels[1]
        return nx, ny
    
    def update_mouse_pos(self, pos):
        if self.isInvalidPosition(pos):
            self.clearMouse()
            return
        else:
            nx, ny = self.relativepos(pos)
            if self.mpos:
                if self.mpos != (nx, ny):
                    self.release_mouse(pos, clear=False)
                self.mpos = nx, ny
                self.wall.set_selected(nx, ny)
            else:
                self.wall.set_highlight(nx, ny)
        
            
    def draw_background(self, screen):
        screen.fill((0,0,0))
        
    def draw_stage(self, screen):
        time =  (pygame.time.get_ticks() - self.start) / 1000.0
        self.offset = screen_size[0] - int(time * self.speed)
        self.wall.draw(screen, self.offset)
"""    
    def draw_water(self, screen):
        column = block_dimensions[0] - self.offset / block_dimensions[0]
        offset = - (abs(self.offset - screen_size[0]) % block_pixels[0])
        def fill_water(column, previous):
            def isChild(pos):
                x, y = pos
                if (x < column or x >= len(self.wall) or
                    y < 0 or y >= block_dimensions[1] or
                    (x, y) in previous):
                    return False
                else:
                    return True
            x, y = pos
            a = (x + 1, y)
            b = (x - 1, y)
            c = (x, y + 1)
            d = (x, y - 1)
            filtered = [i for i in (a, b, c, d) if isChild(i)]
            for i in filtered:
                x, y = i
                
            
        self.fill_water(column)
        for i in range(0, block_dimensions[0] + 1):
            for j in range(block_dimensions[1]):
                if self.water[i + column][j]:
                    screen.fill((0, 0, 255), ((i * block_pixels[0] + offset, j * block_pixels[1]), block_pixels))
"""

if __name__ == '__main__':
    main()